==> axual-client-python-1.1.1.tar.gz: axual-client-python-1.1.1/axualclient/serializing_producer.py <==
# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import threading
from time import sleep
from typing import List, Callable

import confluent_kafka
from confluent_kafka import SerializingProducer as KafkaSerializingProducer, TopicPartition
from confluent_kafka.admin import ClusterMetadata
from confluent_kafka.serialization import Serializer

from axualclient.avro import AvroSerializer
from axualclient.discovery import DiscoveryClient, BOOTSTRAP_SERVERS_KEY, TIMESTAMP_KEY, DiscoveryClientRegistry, \
    SCHEMA_REGISTRY_URL_KEY, DISTRIBUTOR_TIMEOUT_KEY, DISTRIBUTOR_DISTANCE_KEY
from axualclient.patterns import resolve_topic
from axualclient.util import first_of_string_list, filter_axual_configuration, calculate_producer_switch_timeout, \
    is_documented_by

logger = logging.getLogger(__name__)


class SerializingProducer(DiscoveryClient):

    def __init__(self,
                 configuration: dict,
                 *args, **kwargs):
        """
        Instantiate an AVRO producer for Axual. The _producer attribute is
         the confluent_kafka SerializingProducer class.

        Parameters
        ----------
        configuration: dict
            App config information
        *args and **kwargs:
            Other parameters that can be passed to confluent_kafka Producer.

        """
        self._producer = None                    # has not received discovery result yet

        self.init_args = args
        self.init_kwargs = kwargs

        self.configuration = configuration
        # bootstrap servers & key/value serializers are not available at this point yet
        self.configuration['security.protocol'] = 'SSL'
        if 'partitioner' not in self.configuration.keys():
            self.configuration['partitioner'] = 'murmur2_random'

        self.switch_lock = threading.Lock()
        self.init_lock = threading.Lock()

        self.discovery_result = {}
        self.discovery_fetcher = DiscoveryClientRegistry.register_client(
            self.configuration, self
        )

    def wait_for_initialization(self) -> None:
        if self._producer is not None:
            return
        with self.init_lock:
            self.discovery_fetcher.wait_for_discovery_result()

    def _do_with_switch_lock(self, func):
        self.wait_for_initialization()
        with self.switch_lock:
            return func()

    def _get_producer(self):
        """ Convenience method added such that lambda expressions do not get evaluated on the old proxied object """
        return self._producer

    def on_discovery_properties_changed(self, discovery_result: dict) -> None:
        """ A new discovery result has been received, need to switch """
        with self.switch_lock:
            self.discovery_result = discovery_result
            # plug in the new bootstrap servers
            self.configuration['bootstrap.servers'] = discovery_result[BOOTSTRAP_SERVERS_KEY]

            # initialize serializers against new schema registry
            sr_url = first_of_string_list(discovery_result[SCHEMA_REGISTRY_URL_KEY])

            self._reconfigure_serializer(self.configuration['key.serializer'], sr_url)
            self._reconfigure_serializer(self.configuration['value.serializer'], sr_url)

            # Switch producer
            if self._producer is not None:
                self._producer.flush()
                switch_timeout = calculate_producer_switch_timeout(
                    self._is_keeping_order(),
                    int(discovery_result[DISTRIBUTOR_TIMEOUT_KEY]),
                    int(discovery_result[DISTRIBUTOR_DISTANCE_KEY]),
                    discovery_result[TIMESTAMP_KEY]
                )
                sleep(switch_timeout)

            kafka_properties = filter_axual_configuration(self.configuration)
            self._producer = KafkaSerializingProducer(
                kafka_properties, *self.init_args, **self.init_kwargs
            )

    def _is_keeping_order(self):
        return self.configuration.get('max.in.flight.requests.per.connection') in [1, '1']

    def _reconfigure_serializer(self, serializer: Serializer, schema_registry_url: str) -> None:
        """
        AvroSerializers need to be reconfigured with the new target SchemaRegistry.
        @param serializer: the serializer to reconfigure
        @param schema_registry_url: the new SchemaRegistry url
        @return: None
        """
        if isinstance(serializer, AvroSerializer):
            serializer.configure({
                'url': schema_registry_url,
                'ssl.ca.location': self.configuration['ssl.ca.location'],
                'ssl.key.location': self.configuration['ssl.key.location'],
                'ssl.certificate.location': self.configuration['ssl.certificate.location']
            })

    @is_documented_by(confluent_kafka.cimpl.Producer.produce)
    def produce(self,
                topic: str,
                value=None,
                key=None,
                partition: int = None,
                on_delivery: Callable = None,
                timestamp: int = None,
                headers: dict = None,
                *args, **kwargs) -> None:
        if value is not None:
            kwargs['value'] = value
        if key is not None:
            kwargs['key'] = key
        if partition is not None:
            kwargs['partition'] = partition
        if on_delivery is not None:
            kwargs['on_delivery'] = on_delivery
        if timestamp is not None:
            kwargs['timestamp'] = timestamp
        if headers is not None:
            kwargs['headers'] = headers
        self._do_with_switch_lock(
            lambda:
            self._get_producer().produce(resolve_topic(self.discovery_result, topic), *args, **kwargs)
        )

    # producer interface
    @is_documented_by(confluent_kafka.cimpl.Producer.abort_transaction)
    def abort_transaction(self, timeout: float = None, *args, **kwargs):
        if timeout is not None:
            self._do_with_switch_lock(
                lambda: self._get_producer().abort_transaction(timeout, *args, **kwargs)
            )
        else:
            self._do_with_switch_lock(
                lambda: self._get_producer().abort_transaction(*args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.begin_transaction)
    def begin_transaction(self, *args, **kwargs):
        self._do_with_switch_lock(
            lambda: self._get_producer().begin_transaction(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.commit_transaction)
    def commit_transaction(self, timeout: float = None, *args, **kwargs):
        if timeout is not None:
            self._do_with_switch_lock(
                lambda: self._get_producer().commit_transaction(timeout, *args, **kwargs)
            )
        else:
            self._do_with_switch_lock(
                lambda: self._get_producer().commit_transaction(*args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.flush)
    def flush(self, timeout: float = None, *args, **kwargs) -> int:
        if timeout is None:
            return self._do_with_switch_lock(
                lambda: self._get_producer().flush(*args, **kwargs)
            )
        return self._do_with_switch_lock(
            lambda: self._get_producer().flush(timeout, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.init_transactions)
    def init_transactions(self, timeout: float, *args, **kwargs) -> None:
        self._do_with_switch_lock(
            lambda: self._get_producer().init_transactions(timeout, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.list_topics)
    def list_topics(self,
                    topic: str = None,
                    timeout: float = -1,
                    *args, **kwargs) -> ClusterMetadata:
        if topic is not None:
            return self._do_with_switch_lock(
                lambda: self._get_producer().list_topics(topic, timeout, *args, **kwargs)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_producer().list_topics(timeout, *args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.poll)
    def poll(self, timeout=None, *args, **kwargs):
        if timeout is not None:
            kwargs['timeout'] = timeout
        return self._do_with_switch_lock(
            lambda: self._get_producer().poll(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.purge)
    def purge(self, in_queue: bool = True, in_flight: bool = True, blocking: bool = True, *args, **kwargs):
        return self._do_with_switch_lock(
            lambda: self._get_producer().purge(in_queue, in_flight, blocking, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.send_offsets_to_transaction)
    def send_offsets_to_transaction(self,
                                    offsets: List[TopicPartition],
                                    group_metadata: object,
                                    timeout=None,
                                    *args, **kwargs):
        if timeout is not None:
            kwargs['timeout'] = timeout
        return self._do_with_switch_lock(
            lambda: self._get_producer().send_offsets_to_transaction(offsets, group_metadata, *args, **kwargs)
        )

    # obj interface
    def __class__(self, *args, **kwargs):
        return self._producer.__class__(*args, **kwargs)

    def __dict__(self, *args, **kwargs):
        return self._producer.__dict__(*args, **kwargs)

    def __dir__(self, *args, **kwargs):
        return self._producer.__dir__(*args, **kwargs)

    def __doc__(self, *args, **kwargs):
        return self._producer.__doc__(*args, **kwargs)

    def __eq__(self, *args, **kwargs):
        return self._producer.__eq__(*args, **kwargs)

    def __format__(self, *args, **kwargs):
        return self._producer.__format__(*args, **kwargs)

    def __ge__(self, *args, **kwargs):
        return self._producer.__ge__(*args, **kwargs)

    def __gt__(self, *args, **kwargs):
        return self._producer.__gt__(*args, **kwargs)

    def __hash__(self, *args, **kwargs):
        return self._producer.__hash__(*args, **kwargs)

    def __le__(self, *args, **kwargs):
        return self._producer.__le__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self._producer.__len__(*args, **kwargs)

    def __lt__(self, *args, **kwargs):
        return self._producer.__lt__(*args, **kwargs)

    def __module__(self, *args, **kwargs):
        return self._producer.__module__(*args, **kwargs)

    def __ne__(self, *args, **kwargs):
        return self._producer.__ne__(*args, **kwargs)

    def __reduce__(self, *args, **kwargs):
        return self._producer.__reduce__(*args, **kwargs)

    def __reduce_ex__(self, *args, **kwargs):
        return self._producer.__reduce_ex__(*args, **kwargs)

    def __repr__(self, *args, **kwargs):
        return self._producer.__repr__(*args, **kwargs)

    def __sizeof__(self, *args, **kwargs):
        return self._producer.__sizeof__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        return self._producer.__str__(*args, **kwargs)

    def __weakref__(self, *args, **kwargs):
        return self._producer.__weakref__(*args, **kwargs)

==> axual-client-python-1.1.5.tar.gz: axual-client-python-1.1.5/axualclient/serializing_producer.py <==
# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from confluent_kafka import SerializingProducer as KafkaSerializingProducer
from confluent_kafka.serialization import Serializer

from axualclient.avro import AvroSerializer
from axualclient.axual_producer import _AxualProducer
from axualclient.discovery import BOOTSTRAP_SERVERS_KEY, SCHEMA_REGISTRY_URL_KEY
from axualclient.util import first_of_string_list, filter_axual_configuration

logger = logging.getLogger(__name__)


class SerializingProducer(_AxualProducer):

    def __init__(self,
                 configuration: dict,
                 *args, **kwargs):
        """
        Instantiate an AVRO producer for Axual. The _producer attribute is
         the confluent_kafka SerializingProducer class.

        Parameters
        ----------
        configuration: dict
            App config information
        *args and **kwargs:
            Other parameters that can be passed to confluent_kafka Producer.

        """
        super().__init__(configuration, *args, **kwargs)

    def on_discovery_properties_changed(self, discovery_result: dict) -> None:
        """ A new discovery result has been received, need to switch """
        with self.switch_lock:
            self.discovery_result = discovery_result
            # plug in the new bootstrap servers
            self.configuration['bootstrap.servers'] = discovery_result[BOOTSTRAP_SERVERS_KEY]

            # initialize serializers against new schema registry
            sr_url = first_of_string_list(discovery_result[SCHEMA_REGISTRY_URL_KEY])

            self._reconfigure_serializer(self.configuration['key.serializer'], sr_url)
            self._reconfigure_serializer(self.configuration['value.serializer'], sr_url)

            # Switch producer
            kafka_properties = filter_axual_configuration(self.configuration)
            producer = KafkaSerializingProducer(kafka_properties, *self.init_args, **self.init_kwargs)
            self._switch_producer(discovery_result, producer)

    def _reconfigure_serializer(self, serializer: Serializer, schema_registry_url: str) -> None:
        """
        AvroSerializers need to be reconfigured with the new target SchemaRegistry.
        @param serializer: the serializer to reconfigure
        @param schema_registry_url: the new SchemaRegistry url
        @return: None
        """
        if isinstance(serializer, AvroSerializer):
            serializer.configure({
                'url': schema_registry_url,
                'ssl.ca.location': self.configuration['ssl.ca.location'],
                'ssl.key.location': self.configuration['ssl.key.location'],
                'ssl.certificate.location': self.configuration['ssl.certificate.location']
            })

    def __class__(self, *args, **kwargs):
        return self._producer.__class__(*args, **kwargs)

    def __dir__(self, *args, **kwargs):
        return self._producer.__dir__(*args, **kwargs)

    def __doc__(self, *args, **kwargs):
        return self._producer.__doc__(*args, **kwargs)
